from coalib.misc.DictUtilities import inverse_dicts


def show_bears(local_bears, global_bears, show_bears_callback):
    """
    Extracts all the bears from each enabled section or the sections in the
    targets and passes a dictionary to the show_bears_callback method.

    :param local_bears:         Dictionary of local bears with section names
                                as keys and bear list as values.
    :param global_bears:        Dictionary of global bears with section
                                names as keys and bear list as values.
    :param show_bears_callback: The callback that is used to print these
                                bears. It will get one parameter holding
                                bears as key and the list of section names
                                where it's used as values.
    """
    bears = inverse_dicts(local_bears, global_bears)

    show_bears_callback(bears)



import os
import sys

from coalib.misc.i18n import _

# We use circleci for doing dev releases continuously from master to pypi
build_num = os.getenv('CIRCLE_BUILD_NUM')
if build_num is None:  # pragma: no cover
    build_num = 0

version = (0, 1, 1, "dev"+str(build_num))
version_str = ".".join(str(part) for part in version)


def assert_supported_version():  # pragma: no cover
    if sys.version_info < (3, 2):
        print(_("coala supports only python 3.2 or later."))
        exit(1)



# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from coalib.output.ConsoleInteractor import ConsoleInteractor
from coalib.output.printers.ConsolePrinter import ConsolePrinter
from coalib.misc.StringConstants import StringConstants
from coalib.processes.Processing import execute_section
from coalib.settings.ConfigurationGathering import gather_configuration
from coalib.output.ShowBears import show_bears
from coalib.misc.i18n import _


def main():
    log_printer = ConsolePrinter()
    interactor = ConsoleInteractor(log_printer)
    exitcode = 0
    try:
        did_nothing = True
        yielded_results = False
        (sections,
         local_bears,
         global_bears,
         targets) = gather_configuration(interactor.acquire_settings,
                                         log_printer)

        if bool(sections["default"].get("show_bears", "False")):
            show_bears(local_bears,
                       global_bears,
                       interactor.show_bears)
            did_nothing = False
        else:
            for section_name in sections:
                section = sections[section_name]
                if not section.is_enabled(targets):
                    continue

                interactor.begin_section(section)
                results = execute_section(
                    section=section,
                    global_bear_list=global_bears[section_name],
                    local_bear_list=local_bears[section_name],
                    print_results=interactor.print_results,
                    log_printer=log_printer)
                yielded_results = yielded_results or results[0]
                interactor.finalize(results[3])
                did_nothing = False

        if did_nothing:
            interactor.did_nothing()

        if yielded_results:
            exitcode = 1
    except KeyboardInterrupt:  # Ctrl+C
        print(_("Program terminated by user."))
        exitcode = 130
    except EOFError:  # Ctrl+D
        print(_("Found EOF. Exiting gracefully."))
    except SystemExit as exception:
        exitcode = exception.code
    except Exception as exception:
        UNKNOWN_ERROR = _("An unknown error occurred.") + " " + \
                        StringConstants.THIS_IS_A_BUG
        DESCRIPTION = _("During execution of coala an exception was raised. "
                        "This should never happen. When asked for, the "
                        "following information may help investigating:")

        log_printer.log_exception(UNKNOWN_ERROR + " " + DESCRIPTION,
                                  exception)
        exitcode = 255

    return exitcode



import shutil

from coalib.bearlib.abstractions.SectionCreatable import SectionCreatable
from coalib.results.Result import Result
from coalib.output.printers.Printer import Printer
from coalib.misc.i18n import _
from coalib.settings.FunctionMetadata import FunctionMetadata


class Interactor(SectionCreatable, Printer):
    def __init__(self, log_printer):
        SectionCreatable.__init__(self)
        Printer.__init__(self)
        self.log_printer = log_printer
        self.file_diff_dict = {}
        self.current_section = None

    def _print_result(self, result):
        """
        Prints the result.
        """
        raise NotImplementedError

    def _print_actions(self, actions):
        """
        Prints the given actions and lets the user choose.

        :param actions: A list of FunctionMetadata objects.
        :return:        A touple with the name member of the FunctionMetadata
                        object chosen by the user and a Section containing at
                        least all needed values for the action. If the user did
                        choose to do nothing, return (None, None).
        """
        raise NotImplementedError

    def _print_action_failed(self, action_name, exception):
        """
        Prints out the information that the chosen action failed.

        :param action_name: The name of the action that failed.
        :param exception:   The exception with which it failed.
        """
        raise NotImplementedError

    def print_result(self, result, file_dict):
        """
        Prints the result appropriate to the output medium.

        :param result:    A derivative of Result.
        :param file_dict: A dictionary containing all files with filename as
                          key.
        """
        if not isinstance(result, Result):
            self.log_printer.warn(_("One of the results can not be printed "
                                    "since it is not a valid derivative of "
                                    "the coala result class."))
            return

        self._print_result(result)

        actions = result.get_actions()
        if actions == []:
            return

        action_dict = {}
        metadata_list = []
        for action in actions:
            metadata = action.get_metadata()
            action_dict[metadata.name] = action
            metadata_list.append(metadata)

        # User can always choose no action which is guaranteed to succeed
        while not self.apply_action(metadata_list,
                                    action_dict,
                                    result,
                                    file_dict):
            pass

    def apply_action(self,
                     metadata_list,
                     action_dict,
                     result,
                     file_dict):
        action_name, section = self._print_actions(metadata_list)
        if action_name is None:
            return True

        chosen_action = action_dict[action_name]
        try:
            chosen_action.apply_from_section(result,
                                             file_dict,
                                             self.file_diff_dict,
                                             section)
        except Exception as exception:
            self._print_action_failed(action_name, exception)
            return False

        return True

    def print_results(self, result_list, file_dict):
        """
        Prints all given results. They will be sorted.

        :param result_list: List of the results
        :param file_dict:   Dictionary containing filename: file_contents
        """
        if not isinstance(result_list, list):
            raise TypeError("result_list should be of type list")
        if not isinstance(file_dict, dict):
            raise TypeError("file_dict should be of type dict")

        sorted_results = sorted(result_list)
        for result in sorted_results:
            self.print_result(result, file_dict)

    def acquire_settings(self, settings):
        """
        This method prompts the user for the given settings.

        :param settings: a dictionary with the settings name as key and a list
                         containing a description in [0] and the name of the
                         bears who need this setting in [1] and following.
                         Example:
        {"UseTabs": ["describes whether tabs should be used instead of spaces",
                     "SpaceConsistencyBear",
                     "SomeOtherBear"]}

        :return:         a dictionary with the settings name as key and the
                         given value as value.
        """
        raise NotImplementedError

    def finalize(self, file_dict):
        """
        To be called after all results are given to the interactor.
        """
        for filename in self.file_diff_dict:
            diff = self.file_diff_dict[filename]
            file_dict[filename] = diff.apply(file_dict[filename])

            # Backup original file, override old backup if needed
            shutil.copy2(filename, filename + ".orig")

            # Write new contents
            with open(filename, mode='w') as file:
                file.writelines(file_dict[filename])

    @classmethod
    def get_metadata(cls):
        return FunctionMetadata.from_function(cls.__init__,
                                              ["self", "log_printer"])

    def begin_section(self, section):
        """
        Will be called before the results for a section come in (via
        print_results).

        :param section: The section that will get executed now.
        """
        self.file_diff_dict = {}
        self.current_section = section
        self._print_section_beginning(section)

    def _print_section_beginning(self, section):
        """
        Will be called after initialization current_section in
        begin_section()

        :param section: The section that will get executed now.
        """
        raise NotImplementedError

    def show_bears(self, bears):
        """
        It presents the bears to the user and information about each bear.

        :param bears: A dictionary containing bears as keys and a list of
                      sections which the bear belongs as the value.
        """
        raise NotImplementedError

    def did_nothing(self):
        """
        Will be called after processing a coafile when nothing had to be done,
        i.e. no section was enabled/targeted.
        """
        raise NotImplementedError



import multiprocessing
import queue
import os
import platform
import subprocess

from coalib.collecting.Collectors import collect_files
from coalib.collecting import Dependencies
from coalib.output.printers import LOG_LEVEL
from coalib.processes.BearRunning import run
from coalib.processes.CONTROL_ELEMENT import CONTROL_ELEMENT
from coalib.results.HiddenResult import HiddenResult
from coalib.settings.Setting import path_list
from coalib.misc.i18n import _
from coalib.processes.LogPrinterThread import LogPrinterThread


def get_cpu_count():
    try:
        return multiprocessing.cpu_count()
    # cpu_count is not implemented for some CPU architectures/OSes
    except NotImplementedError:  # pragma: no cover
        return 2


def fill_queue(queue_fill, any_list):
    """
    Takes element from a list and populates a queue with those elements.

    :param queue_fill: The queue to be filled.
    :param any_list:   List containing the elements.
    """
    for elem in any_list:
        queue_fill.put(elem)


def get_running_processes(processes):
    return sum((1 if process.is_alive() else 0) for process in processes)


def create_process_group(command_array, **kwargs):
    if platform.system() == "Windows":  # pragma: no cover
        p = subprocess.Popen(command_array,
                             creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
                             **kwargs)
    else:
        p = subprocess.Popen(command_array,
                             preexec_fn=os.setsid,
                             **kwargs)
    return p

def print_result(results,
                 file_dict,
                 retval,
                 print_results):
    """
    Takes the results produced by each bear and gives them to the interactor to
    present to the user.

    :param results:       A list of results.
    :param file_dict:     A dictionary containing the name of files and its
                          contents.
    :param retval:        It is True if no results were yielded ever before.
                          If it is False this function will return False no
                          matter what happens. Else it depends on if this
                          invocation yields results.
    :param print_results: Prints all given results appropriate to the output
                          medium.
    :return:              Returns False if any results were yielded. Else True.
    """
    results = list(filter(lambda result: not isinstance(result, HiddenResult),
                          results))
    print_results(results, file_dict)
    return retval or len(results) > 0


def get_file_dict(filename_list, log_printer):
    """
    Reads all files into a dictionary.

    :param filename_list: List of names of paths to files to get contents of.
    :param log_printer:   The logger which logs errors.
    :return:              Reads the content of each file into a dictionary
                          with filenames as keys.
    """
    file_dict = {}
    for filename in filename_list:
        try:
            with open(filename, "r", encoding="utf-8") as f:
                file_dict[filename] = f.readlines()
        except UnicodeDecodeError:
            log_printer.warn(_("Failed to read file '{}'. It seems to contain "
                               "non-unicode characters. Leaving it "
                               "out.".format(filename)))
        except Exception as exception:  # pragma: no cover
            log_printer.log_exception(_("Failed to read file '{}' because of "
                                        "an unknown error. Leaving it "
                                        "out.").format(filename),
                                      exception,
                                      log_level=LOG_LEVEL.WARNING)

    return file_dict


def instantiate_bears(section,
                      local_bear_list,
                      global_bear_list,
                      file_dict,
                      message_queue):
    """
    Instantiates each bear with the arguments it needs.

    :param section:          The section the bears belong to.
    :param local_bear_list:  List of local bears to instantiate.
    :param global_bear_list: List of global bears to instantiate.
    :param file_dict:        Dictionary containing filenames and their
                             contents.
    :param message_queue:    Queue responsible to maintain the messages
                             delivered by the bears.
    """
    for i in range(len(local_bear_list)):
        local_bear_list[i] = local_bear_list[i](section,
                                                message_queue,
                                                TIMEOUT=0.1)
    for i in range(len(global_bear_list)):
        global_bear_list[i] = global_bear_list[i](file_dict,
                                                  section,
                                                  message_queue,
                                                  TIMEOUT=0.1)


def instantiate_processes(section,
                          local_bear_list,
                          global_bear_list,
                          job_count,
                          log_printer):
    """
    Instantiate the number of processes that will run bears which will be
    responsible for running bears in a multiprocessing environment.

    :param section:          The section the bears belong to.
    :param local_bear_list:  List of local bears belonging to the section.
    :param global_bear_list: List of global bears belonging to the section.
    :param job_count:        Max number of processes to create.
    :param log_printer:      The log printer to warn to.
    :return:                 A tuple containing a list of processes,
                             and the arguments passed to each process which are
                             the same for each object.
    """
    filename_list = collect_files(path_list(section.get('files', "")))
    file_dict = get_file_dict(filename_list, log_printer)

    manager = multiprocessing.Manager()
    global_bear_queue = multiprocessing.Queue()
    filename_queue = multiprocessing.Queue()
    local_result_dict = manager.dict()
    global_result_dict = manager.dict()
    message_queue = multiprocessing.Queue()
    control_queue = multiprocessing.Queue()

    bear_runner_args = {"file_name_queue": filename_queue,
                        "local_bear_list": local_bear_list,
                        "global_bear_list": global_bear_list,
                        "global_bear_queue": global_bear_queue,
                        "file_dict": file_dict,
                        "local_result_dict": local_result_dict,
                        "global_result_dict": global_result_dict,
                        "message_queue": message_queue,
                        "control_queue": control_queue,
                        "timeout": 0.1}

    instantiate_bears(section,
                      local_bear_list,
                      global_bear_list,
                      file_dict,
                      message_queue)
    fill_queue(filename_queue, file_dict.keys())
    fill_queue(global_bear_queue, range(len(global_bear_list)))

    return ([multiprocessing.Process(target=run, kwargs=bear_runner_args)
             for i in range(job_count)],
            bear_runner_args)


def process_queues(processes,
                   control_queue,
                   local_result_dict,
                   global_result_dict,
                   file_dict,
                   print_results):
    """
    Iterate the control queue and send the results recieved to the interactor
    so that they can be presented to the user.

    :param processes:          List of processes which can be used to run
                               Bears.
    :param control_queue:      Containing control elements that indicate
                               whether there is a result available and which
                               bear it belongs to.
    :param local_result_dict:  Dictionary containing results respective to
                               local bears. It is modified by the processes
                               i.e. results are added to it by multiple
                               processes.
    :param global_result_dict: Dictionary containing results respective to
                               global bears. It is modified by the processes
                               i.e. results are added to it by multiple
                               processes.
    :param file_dict:          Dictionary containing file contents with
                               filename as keys.
    :param print_results:      Prints all given results appropriate to the
                               output medium.
    :return:                   Return True if all bears execute succesfully and
                               Results were delivered to the user. Else False.
    """
    running_processes = get_running_processes(processes)
    retval = False
    # Number of processes working on local bears
    local_processes = len(processes)
    global_result_buffer = []

    # One process is the logger thread
    while local_processes > 1 and running_processes > 1:
        try:
            control_elem, index = control_queue.get(timeout=0.1)

            if control_elem == CONTROL_ELEMENT.LOCAL_FINISHED:
                local_processes -= 1
            elif control_elem == CONTROL_ELEMENT.LOCAL:
                assert local_processes != 0
                retval = print_result(local_result_dict[index],
                                      file_dict,
                                      retval,
                                      print_results)
            elif control_elem == CONTROL_ELEMENT.GLOBAL:
                global_result_buffer.append(index)
        except queue.Empty:
            running_processes = get_running_processes(processes)

    # Flush global result buffer
    for elem in global_result_buffer:
        retval = print_result(global_result_dict[elem],
                              file_dict,
                              retval,
                              print_results)

    running_processes = get_running_processes(processes)
    # One process is the logger thread
    while running_processes > 1:
        try:
            control_elem, index = control_queue.get(timeout=0.1)

            if control_elem == CONTROL_ELEMENT.GLOBAL:
                retval = print_result(global_result_dict[index],
                                      file_dict,
                                      retval,
                                      print_results)
            else:
                assert control_elem == CONTROL_ELEMENT.GLOBAL_FINISHED
                running_processes = get_running_processes(processes)

        except queue.Empty:
            running_processes = get_running_processes(processes)

    return retval


def execute_section(section,
                    global_bear_list,
                    local_bear_list,
                    print_results,
                    log_printer):
    """
    Executes the section with the given bears.

    The execute_section method does the following things:
    1. Prepare a Process
      * Load files
      * Create queues
    2. Spawn up one or more Processes
    3. Output results from the Processes
    4. Join all processes

    :param section:          The section to execute.
    :param global_bear_list: List of global bears belonging to the section.
    :param local_bear_list:  List of local bears belonging to the section.
    :param print_results:    Prints all given results appropriate to the
                             output medium.
    :param log_printer:      The log_printer to warn to.
    :return:                 Tuple containing a bool (True if results were
                             yielded, False otherwise), a Manager.dict
                             containing all local results(filenames are key)
                             and a Manager.dict containing all global bear
                             results (bear names are key).
    """
    local_bear_list = Dependencies.resolve(local_bear_list)
    global_bear_list = Dependencies.resolve(global_bear_list)

    running_processes = get_cpu_count()
    processes, arg_dict = instantiate_processes(section,
                                                local_bear_list,
                                                global_bear_list,
                                                running_processes,
                                                log_printer)

    logger_thread = LogPrinterThread(arg_dict["message_queue"],
                                     log_printer)
    # Start and join the logger thread along with the processes to run bears
    processes.append(logger_thread)

    for runner in processes:
        runner.start()

    try:
        return (process_queues(processes,
                               arg_dict["control_queue"],
                               arg_dict["local_result_dict"],
                               arg_dict["global_result_dict"],
                               arg_dict["file_dict"],
                               print_results),
                arg_dict["local_result_dict"],
                arg_dict["global_result_dict"],
                arg_dict["file_dict"])
    finally:
        logger_thread.running = False

        for runner in processes:
            runner.join()



try:
    # This import has side effects and is needed to make input() behave nicely
    import readline
except ImportError: # pragma: no cover
    pass

from coalib.output.printers.ConsolePrinter import ConsolePrinter
from coalib.results.RESULT_SEVERITY import (
    RESULT_SEVERITY,
    RESULT_SEVERITY_COLORS)
from coalib.output.printers.LOG_LEVEL import LOG_LEVEL
from coalib.output.Interactor import Interactor
from coalib.misc.i18n import _
from coalib.settings.Setting import Setting


class ConsoleInteractor(Interactor, ConsolePrinter):
    STR_GET_VAL_FOR_SETTING = _("Please enter a value for the setting \"{}\" "
                                "({}) needed by {}: ")
    STR_LINE_DOESNT_EXIST = _("The line belonging to the following result "
                              "cannot be printed because it refers to a line "
                              "that doesn't seem to exist in the given file.")
    STR_PROJECT_WIDE = _("Project wide:")
    FILE_NAME_COLOR = "blue"
    FILE_LINES_COLOR = "blue"

    def __init__(self,
                 log_printer,
                 pre_padding: int=3,
                 print_colored=True):
        """
        A ConsoleInteractor uses the Console to interact with the user.

        :param log_printer: The LogPrinter to use for logging.
        :param pre_padding: Number of code lines to show before a result as
                            context.
        """
        Interactor.__init__(self, log_printer=log_printer)
        ConsolePrinter.__init__(self, print_colored=print_colored)

        self.pre_padding = pre_padding

    def acquire_settings(self, settings_names_dict):
        if not isinstance(settings_names_dict, dict):
            raise TypeError("The settings_names_dict parameter has to be a "
                            "dictionary.")

        result = {}
        for setting_name, arr in settings_names_dict.items():
            value = self._require_setting(setting_name, arr)
            if value is not None:
                result[setting_name] = value

        return result

    def _require_setting(self, setting_name, arr):
        if not isinstance(arr, list) or len(arr) < 2:
            self.log_printer.log(LOG_LEVEL.WARNING,
                                 _("One of the given settings ({}) is not "
                                   "properly described.").
                                 format(str(setting_name)))

            return None

        if len(arr) == 2:
            needed = arr[1]
        else:
            needed = ", ".join(arr[1:-1]) + _(" and ") + arr[-1]

        return input(self.STR_GET_VAL_FOR_SETTING.format(str(setting_name),
                                                         str(arr[0]),
                                                         needed))

    def _format_line(self, line, real_nr="", sign="|", mod_nr="", symbol="", ):
        return "|{:>4}{}{:>4}|{:1}{}".format(real_nr,
                                             sign,
                                             mod_nr,
                                             symbol,
                                             line.rstrip("\n"))

    def _print_result(self, result):
        self.print(self._format_line(
            "[{sev}] {bear}:".format(
                sev=RESULT_SEVERITY.__str__(result.severity),
                bear=result.origin)),
            color=RESULT_SEVERITY_COLORS[result.severity])
        self.print(*[self._format_line(line)
                     for line in result.message.split("\n")],
                   delimiter="\n")

    def _print_actions(self, actions):
        choice = self._choose_action(actions)

        if choice == 0:
            return None, None

        return self._get_action_info(actions[choice - 1])

    def _choose_action(self, actions):
        self.print(self._format_line(
            _("The following options are applicable to this result:")))

        while True:
            self.print(self._format_line(" 0: " + _("Do nothing.")))
            for i, action in enumerate(actions):
                self.print(self._format_line("{:>2}: {}".format(i + 1,
                                                                action.desc)))

            try:
                line = self._format_line(_("Please enter the number of the "
                                           "action you want to execute. "))
                choice = int(input(line))
                if 0 <= choice <= len(actions):
                    return choice
            except ValueError:
                pass

            self.print(self._format_line(_("Please enter a valid number.")))

    def _print_action_failed(self, action_name, exception):
        self.log_printer.log_exception("Failed to execute the action "
                                       "{}.".format(action_name),
                                       exception)

    def _get_action_info(self, action):
        params = action.non_optional_params

        if self.current_section is None:
            raise ValueError("current_section has to be intializied.")

        for param_name in params:
            if param_name not in self.current_section:
                question = self._format_line(
                    _("Please enter a value for the parameter '{}' ({}): ")
                    .format(param_name, params[param_name][0]))
                self.current_section.append(Setting(param_name,
                                                    input(question)))

        return action.name, self.current_section

    def _print_segregation(self):
        self.print(self._format_line(line="",
                                     real_nr="...",
                                     sign="|",
                                     mod_nr="..."),
                   color=self.FILE_LINES_COLOR)

    def _print_lines(self, file_dict, current_line, result_line, result_file):
        """
        Prints the lines between the current and the result line. If needed
        they will be shortened.
        """
        line_delta = result_line - current_line

        if line_delta > self.pre_padding:
            self._print_segregation()

            for i in range(max(result_line - self.pre_padding, 1),
                           result_line + 1):
                self.print(
                    self._format_line(
                        line=file_dict[result_file][i - 1],
                        real_nr=i,
                        mod_nr=i),
                    color=self.FILE_LINES_COLOR)
        else:
            for i in range(1, line_delta + 1):
                self.print(
                    self._format_line(
                        line=file_dict[result_file][current_line + i - 1],
                        real_nr=current_line + i,
                        mod_nr=current_line + i),
                    color=self.FILE_LINES_COLOR)

    def print_results(self, result_list, file_dict):
        if not isinstance(result_list, list):
            raise TypeError("result_list should be of type list")
        if not isinstance(file_dict, dict):
            raise TypeError("file_dict should be of type dict")

        # We can't use None since we need line 109 be executed if file of first
        # result is None
        current_file = False
        current_line = 0

        for result in sorted(result_list):
            if result.file != current_file:
                if result.file in file_dict or result.file is None:
                    current_file = result.file
                    current_line = 0
                    self.print("\n\n{}".format(current_file
                                               if current_file is not None
                                               else self.STR_PROJECT_WIDE),
                               color=self.FILE_NAME_COLOR)
                else:
                    self.log_printer.warn(_("A result ({}) cannot be printed "
                                            "because it refers to a file that "
                                            "doesn't seem to "
                                            "exist.").format(str(result)))
                    continue

            if result.line_nr is not None:
                if current_file is None:
                    raise AssertionError("A result with a line_nr should also "
                                         "have a file.")
                if result.line_nr < current_line:  # pragma: no cover
                    raise AssertionError("The sorting of the results doesn't "
                                         "work correctly.")
                if len(file_dict[result.file]) < result.line_nr - 1:
                    self.print(self._format_line(
                        line=self.STR_LINE_DOESNT_EXIST))
                else:
                    self._print_lines(file_dict,
                                      current_line,
                                      result.line_nr,
                                      result.file)
                    current_line = result.line_nr

            self.print_result(result, file_dict)

    def _print_section_beginning(self, section):
        self.print(_("Executing section {name}...").format(name=section.name))

    def show_bears(self, bears):
        """
        Presents all bears being used in a stylized manner.

        :param bears: Its a dictionary with bears as keys and list of sections
                      containing those bears as values.
        """
        if not bears:
            self.print(_("No bears to show."))
        else:
            for bear in sorted(bears.keys(),
                               key=lambda bear: bear.__class__.__name__):
                self._show_bear(bear, bears[bear], bear.get_metadata())

    def _show_enumeration(self, title, items, indentation, no_items_text):
        """
        This function takes as input an iterable object (preferably a list or
        a dict). And prints in a stylized format. If the iterable object is
        empty, it prints a specific statement give by the user. An e.g :

        <indentation>Title:
        <indentation> * Item 1
        <indentation> * Item 2

        :param title:         Title of the text to be printed
        :param items:         The iterable object.
        :param indentation:   Number of spaces to indent every line by.
        :param no_items_text: Text printed when iterable object is empty.
        """
        if not items:
            self.print(indentation + no_items_text)
        else:
            self.print(indentation + title)
            if isinstance(items, dict):
                for key, value in items.items():
                    self.print(indentation + " * " + key + ": " + value[0])
            else:
                for item in items:
                    self.print(indentation + " * " + item)
        self.print()

    def _show_bear(self, bear, sections, metadata):
        self.print("{bear}:".format(bear=bear.__name__))
        self.print("  " + metadata.desc + "\n")

        self._show_enumeration(_("Used in:"),
                               sections,
                               "  ",
                               _("No sections."))
        self._show_enumeration(_("Needed Settings:"),
                               metadata.non_optional_params,
                               "  ",
                               _("No needed settings."))
        self._show_enumeration(_("Optional Settings:"),
                               metadata.optional_params,
                               "  ",
                               _("No optional settings."))

    def did_nothing(self):
        self.print(_("No existent section was targeted or enabled. Nothing "
                     "to do."))



from coalib.output.Interactor import Interactor


class NullInteractor(Interactor):
    """
    This interactor can be used for headless setups, it requires no interaction
    from the user.
    """

    def print_result(self, result, file_dict):
        pass

    def print_results(self, result_list, file_dict):
        pass

    def did_nothing(self):
        pass

    def begin_section(self, section):
        pass

    def acquire_settings(self, settings):
        # We can't get settings here, if bears need them, they better fail.
        pass

    def show_bears(self, bears):
        pass
